/*
    This file is part of the KDE libraries
    SPDX-FileCopyrightText: 2006, 2007 Thomas Braxton <kde.braxton@gmail.com>
    SPDX-FileCopyrightText: 1999 Preston Brown <pbrown@kde.org>
    SPDX-FileCopyrightText: 1997-1999 Matthias Kalle Dalheimer <kalle@kde.org>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#include "kconfigbase.h"

#include "config-kconfig.h"
#include "kconfiggroup.h"

#include <QString>

#if KCONFIG_USE_DBUS
#include <QDBusConnection>
#include <QDBusMessage>
#include <QStringList>
#include <QVector>
#include <QDebug>
#include <QMutex>
#include <QMutexLocker>
#include <unistd.h>
#include <pwd.h>
#include "kconfig.h"
#include "kconfigserialization.h"
#endif // KCONFIG_USE_DBUS

bool KConfigBase::s_notification_warn_once = true;

bool KConfigBase::hasGroup(const QString &group) const
{
    return hasGroupImpl(group.toUtf8());
}

bool KConfigBase::hasGroup(const char *group) const
{
    return hasGroupImpl(QByteArray(group));
}

bool KConfigBase::hasGroup(const QByteArray &group) const
{
    return hasGroupImpl(group);
}

KConfigGroup KConfigBase::group(const QByteArray &b)
{
    return groupImpl(b);
}

KConfigGroup KConfigBase::group(const QString &str)
{
    return groupImpl(str.toUtf8());
}

KConfigGroup KConfigBase::group(const char *str)
{
    return groupImpl(QByteArray(str));
}

const KConfigGroup KConfigBase::group(const QByteArray &b) const
{
    return groupImpl(b);
}

const KConfigGroup KConfigBase::group(const QString &s) const
{
    return groupImpl(s.toUtf8());
}

const KConfigGroup KConfigBase::group(const char *s) const
{
    return groupImpl(QByteArray(s));
}

void KConfigBase::deleteGroup(const QByteArray &group, WriteConfigFlags flags)
{
    deleteGroupImpl(group, flags);
}

void KConfigBase::deleteGroup(const QString &group, WriteConfigFlags flags)
{
    deleteGroupImpl(group.toUtf8(), flags);
}

void KConfigBase::deleteGroup(const char *group, WriteConfigFlags flags)
{
    deleteGroupImpl(QByteArray(group), flags);
}

bool KConfigBase::isGroupImmutable(const QByteArray &aGroup) const
{
    return isGroupImmutableImpl(aGroup);
}

bool KConfigBase::isGroupImmutable(const QString &aGroup) const
{
    return isGroupImmutableImpl(aGroup.toUtf8());
}

bool KConfigBase::isGroupImmutable(const char *aGroup) const
{
    return isGroupImmutableImpl(QByteArray(aGroup));
}

bool KConfigBase::isUpdateNotificationEnabled() const
{
    return isUpdateNotificationEnabledImpl();
}

void KConfigBase::setUpdateNotificationEnabled(bool new_state)
{
    setUpdateNotificationEnabledImpl(new_state);
}

KConfigBase::~KConfigBase()
{
}

KConfigBase::KConfigBase()
{
}

void KConfigBase::virtual_hook(int, void *)
{
}

void KConfigBase::sendNotification(const char *key, const QByteArray &value,
    const QStringList &entry_paths, const QStringList &groups,
    const QString &filename, QStandardPaths::StandardLocation location_type,
    WriteConfigFlags flags, EntryNotificationType notification_type)
{
#if KCONFIG_USE_DBUS
    {
        static QMutex s_globalsInitMutex;
        static bool s_globalsInitialized = false;
        static bool s_notificationsGloballyEnabled;

        { // lock
            QMutexLocker globalsInitMutexLock(&s_globalsInitMutex);

            if (!s_globalsInitialized)
            {
                KConfig globals_config(QStringLiteral("kdeglobals"), KConfig::NoGlobals);
                KConfigGroup alt_group = globals_config.group(QStringLiteral("ALT"));
                s_notificationsGloballyEnabled = alt_group.readEntry(QStringLiteral("EnableDbusNotifications"), false);
                s_globalsInitialized = true;
            }
        } // unlock

        if (!s_notificationsGloballyEnabled)
        {
            return;
        }
    }

    if (!QDBusConnection::sessionBus().isConnected())
    {
        if (s_notification_warn_once)
        {
            qDebug() << QStringLiteral(
                "Cannot connect to the D-Bus session bus.\n"
                "Please check your system settings and try again.\n");
            s_notification_warn_once = false;
        }

        return;
    }

    QStringList message_values_list;

    message_values_list << QStringLiteral("version=1");

    switch (notification_type)
    {
    case EntryNotificationType::Entry:
        message_values_list << QStringLiteral("event=write_entry");
        break;

    case EntryNotificationType::PathEntry:
        message_values_list << QStringLiteral("event=write_path_entry");
        break;

    case EntryNotificationType::PathArrayEntry:
        message_values_list << QStringLiteral("event=write_path_array_entry");
        break;

    case EntryNotificationType::DeleteEntry:
        message_values_list << QStringLiteral("event=delete_entry");
        break;

    case EntryNotificationType::DeleteGroup:
        message_values_list << QStringLiteral("event=delete_group");
        break;
    }

    {
        qint64 initlen = sysconf(_SC_GETPW_R_SIZE_MAX);

        struct passwd result;
        struct passwd *resultp;

        QVector<char> buffer;
        buffer.resize((initlen == -1) ? 1024 : initlen);

        uid_t uid = geteuid();

        int e;

        while ((e = getpwuid_r(uid, &result, buffer.data(), buffer.size(), &resultp)) == ERANGE)
        {
            buffer.resize(buffer.size() * 2);
        }

        if (e != 0)
        {
            qDebug() << QStringLiteral("Failed to serialize config entry: getpwuid_r function failed");
            return;
        }

        if (resultp != nullptr)
        {
            message_values_list << QStringLiteral("username=%1").arg(QString::fromLocal8Bit(resultp->pw_name));
        }

        message_values_list << QStringLiteral("uid=%1").arg(uid);
    }

    message_values_list << QStringLiteral("filename=%1").arg(filename);

    {
        bool ok = false;
        QString location_type_string = serializeStandardLocation(location_type, ok);
        if (!ok)
        {
            qDebug() << QStringLiteral("Failed to serialize config entry: failed to serialize QStandardPaths::StandardLocation enum");
            return;
        }
        message_values_list << QStringLiteral("location_type=%1").arg(location_type_string);
    }

    {
        bool ok = false;
        QString flags_string = serializeWriteConfigFlags(flags, ok);
        if (!ok)
        {
            qDebug() << QStringLiteral("Failed to serialize config entry: failed to serialize KConfigBase::WriteConfigFlags");
            return;
        }

        message_values_list << QStringLiteral("flags=%1").arg(flags_string);
    }

    message_values_list << QStringLiteral("group_name=%1").arg(groups.join(QLatin1Char('|')));

    switch (notification_type)
    {
    case EntryNotificationType::Entry:
    case EntryNotificationType::PathEntry:
    case EntryNotificationType::PathArrayEntry:
    case EntryNotificationType::DeleteEntry:
        message_values_list << QStringLiteral("entry_name=%1").arg(QString::fromLocal8Bit(key));
        break;

    default:
        break;
    }

    switch (notification_type)
    {
    case EntryNotificationType::Entry:
    case EntryNotificationType::PathEntry:
        message_values_list << QStringLiteral("entry_value=%1").arg(QString::fromLocal8Bit(value.isNull() ? QByteArray("") : value));
        break;

    default:
        break;
    }

    if (notification_type == EntryNotificationType::PathArrayEntry)
    {
        int idx = 1;
        auto iter = entry_paths.begin();
        for (; iter != entry_paths.end(); ++iter, ++idx)
        {
            message_values_list << QStringLiteral("entry_path_%1=%2").arg(idx).arg(*iter);
        }
    }

    QDBusMessage message = QDBusMessage::createSignal(QStringLiteral("/"),
        QStringLiteral("org.alt.kde.kconfig.extra.notify"),
        QStringLiteral("ConfigChanged"));

    message << message_values_list;
    QDBusConnection::sessionBus().send(message);
#endif
}
