/*
    This file is part of the KDE libraries
    SPDX-FileCopyrightText: 2020 Aleksei Nikiforov <darktemplar@basealt.ru>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#include "kconfigserialization.h"

#include <QChar>
#include <QStringList>
#include <QMap>
#include <QMutex>
#include <QMutexLocker>

static const QMap<QStandardPaths::StandardLocation, QString> s_standardLocationEnumValues =
{
    { QStandardPaths::DesktopLocation,       QStringLiteral("DesktopLocation") },
    { QStandardPaths::DocumentsLocation,     QStringLiteral("DocumentsLocation") },
    { QStandardPaths::FontsLocation,         QStringLiteral("FontsLocation") },
    { QStandardPaths::ApplicationsLocation,  QStringLiteral("ApplicationsLocation") },
    { QStandardPaths::MusicLocation,         QStringLiteral("MusicLocation") },
    { QStandardPaths::MoviesLocation,        QStringLiteral("MoviesLocation") },
    { QStandardPaths::PicturesLocation,      QStringLiteral("PicturesLocation") },
    { QStandardPaths::TempLocation,          QStringLiteral("TempLocation") },
    { QStandardPaths::HomeLocation,          QStringLiteral("HomeLocation") },
    { QStandardPaths::AppLocalDataLocation,  QStringLiteral("AppLocalDataLocation") },
    { QStandardPaths::CacheLocation,         QStringLiteral("CacheLocation") },
    { QStandardPaths::GenericDataLocation,   QStringLiteral("GenericDataLocation") },
    { QStandardPaths::RuntimeLocation,       QStringLiteral("RuntimeLocation") },
    { QStandardPaths::ConfigLocation,        QStringLiteral("ConfigLocation") },
    { QStandardPaths::DownloadLocation,      QStringLiteral("DownloadLocation") },
    { QStandardPaths::GenericCacheLocation,  QStringLiteral("GenericCacheLocation") },
    { QStandardPaths::GenericConfigLocation, QStringLiteral("GenericConfigLocation") },
    { QStandardPaths::AppDataLocation,       QStringLiteral("AppDataLocation") },
    { QStandardPaths::AppConfigLocation,     QStringLiteral("AppConfigLocation") },
};

static const QMap<QString, KConfigBase::WriteConfigFlag> s_writeConfigFlagsEnumValues =
{
    { QStringLiteral("Persistent"), KConfigBase::Persistent },
    { QStringLiteral("Global"),     KConfigBase::Global },
    { QStringLiteral("Localized"),  KConfigBase::Localized },
    { QStringLiteral("Notify"),     KConfigBase::Notify },
    { QStringLiteral("Normal"),     KConfigBase::Normal },
};

QString serializeStandardLocation(QStandardPaths::StandardLocation value, bool &ok)
{
    auto iter = s_standardLocationEnumValues.find(value);
    if (iter != s_standardLocationEnumValues.end())
    {
        ok = true;
        return iter.value();
    }
    else
    {
        ok = false;
        return QString();
    }
}

QStandardPaths::StandardLocation deserializeStandardLocation(const QString &value, bool &ok)
{
    static QMap<QString, QStandardPaths::StandardLocation> s_standardLocationEnumDeserializationMap;
    static QMutex s_standardLocationEnumDeserializationMapMutex;

    { // lock
        QMutexLocker mapLock(&s_standardLocationEnumDeserializationMapMutex);

        if (s_standardLocationEnumDeserializationMap.isEmpty())
        {
            for (auto iter = s_standardLocationEnumValues.begin(); iter != s_standardLocationEnumValues.end(); ++iter)
            {
                s_standardLocationEnumDeserializationMap[iter.value()] = iter.key();
            }
        }
    } // unlock

    auto iter = s_standardLocationEnumDeserializationMap.find(value);
    if (iter != s_standardLocationEnumDeserializationMap.end())
    {
        ok = true;
        return iter.value();
    }
    else
    {
        ok = false;
        return QStandardPaths::StandardLocation();
    }
}

QString serializeWriteConfigFlags(const KConfigBase::WriteConfigFlags &value, bool &ok)
{
    QStringList result;
    KConfigBase::WriteConfigFlags reconstructed_value;

    for (auto values_iter = s_writeConfigFlagsEnumValues.begin(); values_iter != s_writeConfigFlagsEnumValues.end(); ++values_iter)
    {
        if (value.testFlag(values_iter.value()))
        {
            result.push_back(values_iter.key());
            reconstructed_value |= values_iter.value();
        }
    }

    ok = (value == reconstructed_value);

    return result.join(QChar::fromLatin1('|'));;
}

KConfigBase::WriteConfigFlags deserializeWriteConfigFlags(const QString &value, bool &ok)
{
    KConfigBase::WriteConfigFlags result;

    QStringList values_list = value.split(QChar::fromLatin1('|'));

    bool no_unknown_values = true;

    for (const auto &value_iter: values_list)
    {
        auto map_iter = s_writeConfigFlagsEnumValues.find(value_iter);
        if (map_iter != s_writeConfigFlagsEnumValues.end())
        {
            result |= map_iter.value();
        }
        else
        {
            no_unknown_values = false;
        }
    }

    ok = no_unknown_values;

    return result;
}
